from vcCommand import *
import re
import os.path
import openpyxl
import vcHelpers.Selection
from string import ascii_letters
from string_templates import *

#Available templates that can be used in the Excel file to indicate desired variable pairs
INTEGER_INDEX_TEMPLATES = ["{i}", "{i+1}", "{i-1}"]

#region Main functions
def first_state():
  '''First state, creates the UI panel when the command is activated'''
  global group, uri_xml_base, base, mode_prop
  global  read_xml_prop, tag_list_path
  global server_ip_prop, server_port_prop
  # REMOVE EXISTING COMMAND PROPERTIES
  # cmd = getCommand()
  # cmdProps = cmd.Properties
  # READ URI
  uri = getCommandPath()
  base, filename = os.path.split(uri)
  base = base.decode('utf8')

  uri_xml_base = os.path.join(base, 'templates/ADSConfigurationBase.xml')
  uri_xml_base = uri_xml_base[8:].decode('utf8')

  # CREATE PROPERTIES
  group = 0
  group2 = 0

  mode_prop = createProperty(VC_STRING, 'Connector', VC_PROPERTY_STEP)
  mode_prop.StepValues = ["Beckhoff ADS", "Siemens S7"]
  mode_prop.Value = mode_prop.StepValues[0]
  mode_prop.OnChanged = mode_changed
  mode_prop.IsVisible = False

  tag_list_path = createProperty(VC_URI, 'Variable pairing Excel file')
  tag_list_path.Group = group2 = group2 + 500
  print("Select Excel file with variable pairing information.")
  #print("Expected column format: A: Name, C: Data Type, D: Logical Address")

  server_ip_prop = createProperty(VC_STRING, 'Server IP-Address')
  server_ip_prop.Group = group2 = group2 + 500
  server_ip_prop.Value = "192.168.0.12"
  server_ip_prop.IsVisible = False
  
  server_port_prop = createProperty(VC_STRING, 'Server Port')
  server_port_prop.Group = group2 = group2 + 500
  server_port_prop.Value = "851"
  server_port_prop.IsVisible = False
  
  read_xml_prop = createProperty(VC_BUTTON, 'Generate Configuration XML')
  read_xml_prop.Group = group2 = group2 + 500
  read_xml_prop.OnChanged = button_clicked

  executeInActionPanel() 


def button_clicked(arg):
  if mode_prop.Value == "Beckhoff ADS":
    ads_main(None)
  elif mode_prop.Value == "Siemens S7":
    s7_main(server_ip_prop)
  else:
    pass


def mode_changed(prop):
  '''Triggered when the mode property is changed'''
  if prop.Value == "Beckhoff ADS":
    pass
  elif prop.Value == "Siemens S7":
    pass


#Wrapper because why not (Passes the IP address to the writer function)
def event_wrapper(server_ip_prop):
  def inner(arg):
    s7_main(server_ip_prop)
  return inner
  

def s7_main(server_ip_prop):
  '''Creates a XML configuration file for S7. Reads the Excel and generates the configuration file'''
  #Validate Excel file path
  if len(tag_list_path.Value) < 8:
    print "{} is empty.".format(tag_list_path.Name)
    return

  #Read template XML
  template_text = ""
  with open (uri_xml_base, "r") as template_file:
    template_text = template_file.read()
  if len(template_text) == 0:
    print 'Could not read the template file.'
    return
  
  print("Reading Excel {}".format(tag_list_path.Value))
  tag_data = read_excel_rows(tag_list_path) #Returns raw text data
  if not tag_data:
    print("Could not find any suitable tag data from file.")
    return
  
  tags = parse_tags(tag_data) #Returns a list of Tag objects
  if not tags:
    print "Failed to extract tags from file."
    return
  
  browsing_tags_text = generate_browsing_tags(tags) #For given tags returns a formatted string to create tags for browsing
  if browsing_tags_text is None:
    print "Failed to generate tags for browsing."
    return
  
  #Generate connections
  sim_to_server_text , server_to_sim_text = pair_S7_tags(tags)
  if sim_to_server_text == "" and server_to_sim_text == "":
    print("Failed to find matching pairs for variables")
    return

  #Formats template configuration base file with IP address, browsing tags, and variable pairs
  configuration_text = populate_configuration_info(template_text, server_ip_prop.Value, browsing_tags_text, tags, sim_to_server_text , server_to_sim_text)

  #Save the results to a file
  success = write_text_to_file(configuration_text)
  if success:
    pass


def ads_main(arg):
  '''Main logic for ADS variable pairing'''
  #Validate Excel file path
  if len(tag_list_path.Value) < 8:
    print "{} is empty.".format(tag_list_path.Name)
    return
  
  if not app.Components:
    print("Simulation msut have valid components to create a configuration.")
    return
  
  #Read template XML
  template_text = ""
  with open (uri_xml_base, "r") as template_file:
    template_text = template_file.read()
  if len(template_text) == 0:
    print 'Could not read the template file.'
    return
  
  #Read input Excel
  print("Reading Excel {}".format(tag_list_path.Value))
  pair_input_data = read_excel_rows(tag_list_path) #Returns a list with raw text data from the rows
  if not pair_input_data:
    print("Could not find any suitable tag data from file.")
    return

  pair_infos = initialize_pair_infos(pair_input_data)
  #Read list of rows (ideal pairing formula from Excel) and return list of suitable variable pairs
  ads_pair_info = attempt_variable_pairing(pair_infos)
  if len(ads_pair_info) == 0:
    print "Failed to determine matching pairs."
    return

  #Generate pairings text based on available simulation variables
  configuration_text = populate_ads_configuration(template_text, ads_pair_info)

  #Save the results to a file
  success = write_text_to_file(configuration_text)
  if success:
    print "Select Import from the Configuration toolbar to select this file."

    #Open the file selection dialog for importing the config
    #load_config()
    pass

#endregion


#region Secondary functions
def read_excel_rows(tag_list_path):
  '''Validate Excel file'''
  row_data = []
  file_path = tag_list_path.Value[8:]
  try:
    wb = openpyxl.load_workbook(file_path)
  except Exception, e:
    print e
    print "Make sure that the Excel file is valid. Close the file if it is currently opened."
    return row_data
  #Iterate through all worksheets and extract row data
  for ws in wb.worksheets:
    for row in ws.rows:
      list_with_values = []
      for cell in row:
        val = cell.value
        if val is not None:
          if type(val) == unicode:
            list_with_values.append(val.encode()) #val.encode()
          elif type(val) == long:
            list_with_values.append(str(int(val)))
        else:
          #print "no val", cell.row, cell.column
          list_with_values.append("")
      if len(list_with_values) > 3:
        row_data.append(list_with_values)
      else:
        print "Skipping invalid row", list_with_values
    break #Only check the first worksheet for now
  return row_data


def parse_tags(tag_data):
  '''#Parse tag_data strings into usable Tag objects'''
  tags = []
  for tag_info in tag_data:
    tag = Tag()
    tag.symbol_name = tag_info[0]
    tag.data_type = tag_info[2]
    tag.memory_location = tag_info[3]
    #Check that everything has a value
    if not all([tag.symbol_name, tag.data_type, tag.memory_location]):
      print "Skipping invalid row:", tag_info
      continue
    
    #Parse memory location into memory area, byte index and bit index
    if "%" not in tag.memory_location:
      print "Invalid memory address. Address must start with %.", tag_info
      continue
    if tag.memory_location[1] in DATA_AREA_DICT.keys():
      tag.data_area = DATA_AREA_DICT[tag.memory_location[1]]
    else:
      print "Invalid memory address. Failed to parse memory location", tag_info
      continue
    match= re.search(r"(\d+)", tag.memory_location[2:])
    if match:
      tag.byte_index = match.group()
    else:
      print "Invalid memory address. Failed to parse byte index", tag_info
      continue
    if tag.memory_location[-2] == ".": #Set bit index to 0 if it is not separated in the memory address
      tag.bit_index = "0"
    else:
      tag.bit_index = tag.memory_location[-1]
    
    #Validate data type
    tag.server_data_type = tag.data_type
    if tag.data_type.upper() in DATA_TYPES: #Convert standard data types to upper case. Don't convert user defined data types
      tag.data_type = tag.data_type.upper()
    else:
      tag.data_type = "Unknown"
    
    #Determine the data type length
    if tag.data_type in WORD_LENGTHS.keys():
      tag.word_length = WORD_LENGTHS[tag.data_type]
    elif "B" in tag.memory_location:
      tag.word_length = "Byte"
    elif "W" in tag.memory_location:
      tag.word_length = "Word"
    elif "D" in tag.memory_location:
      tag.word_length = "DWord"
    else:
      tag.word_length = "Bit"
    tags.append(tag)
  return tags


def generate_browsing_tags(tags):
  '''Generates tags for the browsing functionality so that any modifications can be made manually'''
  item_text = ""
  first_pass = True
  for tag in tags:
    if first_pass:
      first_pass = False
    else:
      item_text += "\n\t\t\t\t\t\t\t" #Add new line and tab formatting between each item
    temp_text = S7_BROWSING_VALUE_ITEM_TEMPLATE
    item_text += temp_text.format(MemoryLocation = tag.memory_location, BitIndex = tag.bit_index, ByteIndex = tag.byte_index,
                     DataArea = tag.data_area, DataType = tag.data_type, SymbolName = tag.symbol_name,
                     WordLength = tag.word_length, ServerDataType = tag.server_data_type)
  return item_text


def pair_S7_tags(tags):
  '''Generate pairing text between S7 Tags and simulation variables'''
  sim_to_server_text = ""
  server_to_sim_text = ""
  sim_to_server_count = 0
  server_to_sim_count = 0
  sim_vars = SimulationVariables()
  for tag in tags:
    #Attempt to find a behaviour / property / servo DoF from the simulation that matches the tag symbol name
    paired_component = find_tag_pair(tag, sim_vars)
    #Check if no pair was found for this tag
    if None in (paired_component.component_object, paired_component.name):
      #print "No pair found for", tag.symbol_name
      continue
    
    #Servo DoFs have different structure in the XML than other behaviors or properties
    if paired_component.dof is None: 
      variable_template = STANDARD_VARIABLE_TEMPLATE
    else:
      variable_template = DOF_VARIABLE_TEMPLATE
      variable_template = variable_template.format(DOF = paired_component.dof)
    #Populate variable_template with the component information
    variable_template = variable_template.format(ComponentName = paired_component.name,
                                                 BehaviourName = paired_component.behaviour, PropertyName = paired_component.property)
    
    #Set access type so that simulation can only write to PLC Inputs
    if tag.data_area == "Input":
      access_type = "ReadAndWrite"
    else:
      access_type = "Read"
    #Populate connection item template with the info
    temp_text = S7_CONNECTION_ITEM_TEMPLATE
    temp_text = temp_text.format(VariableTemplate=variable_template, BitIndex=tag.bit_index, ByteIndex=tag.byte_index, Access = access_type,
                                 DataArea=tag.data_area, DataType=tag.data_type, SymbolName=tag.symbol_name, WordLength=tag.word_length)
    
    #Inputs are sent from simu to server. Outputs and memory tags come from server to simu
    if tag.data_area == "Input":
      sim_to_server_count += 1
      if sim_to_server_text != "": #Add line break and tabs between each text block
        sim_to_server_text +="\n\t\t\t\t\t\t\t" 
      sim_to_server_text += temp_text
    else:
      server_to_sim_count += 1
      if server_to_sim_text != "": #Add line break and tabs between each text block
        server_to_sim_text +="\n\t\t\t\t\t\t\t" 
      server_to_sim_text += temp_text
  
  if sim_to_server_count or server_to_sim_count:
    print("Automatically paired some variables in the configuration file.")
    print ("Simulation to PLC = {}, \tPLC to Sim = {}, \t Total = {}.".format(sim_to_server_count, server_to_sim_count, sim_to_server_count + server_to_sim_count))
  return sim_to_server_text, server_to_sim_text


def find_tag_pair(tag, sim_vars):
  '''Attempts to find a simulation variable that matches the tag symbol name and type
  1. See if the tag should be connected to a servo Dof.
  2. If no match found, attempt to find a match from behaviour of given tag type.
  3. If no match found, expand search of properties of given type.
  4. If still no match, go through all properties and all behaviours once more to see if any names match'''
  pair = PairedComponent()
  matched_dof = None
  matched_behaviour = None
  matched_property = None
  #Reals
  if tag.data_type == "REAL":
    if "_M" in tag.symbol_name: #Look for servo DoF
      matched_dof = dof_search_helper(sim_vars, tag.symbol_name)
      # if matched_dof is not None:
      #   print "FOUND DOF", matched_dof.Controller.Component.Name, matched_dof.Controller.Name, matched_dof.Name
    if matched_dof is None:
      matched_behaviour = name_iteration_helper(sim_vars.real_signals, tag.symbol_name)
      if matched_behaviour is None:
        matched_property = name_iteration_helper(sim_vars.real_properties, tag.symbol_name)
  #Booleans
  elif tag.data_type == "BOOL" or tag.word_length == "Bit": #word_length is used to match unknown data types
    matched_behaviour = name_iteration_helper(sim_vars.boolean_signals, tag.symbol_name)
    if matched_behaviour is None:
      matched_property = name_iteration_helper(sim_vars.boolean_properties, tag.symbol_name)
  #Integers
  elif tag.data_type in ("BYTE", "WORD", "DWORD", "SINT", "USINT", "INT", "UINT", "DINT", "UDINT"):
    matched_behaviour = name_iteration_helper(sim_vars.integer_signals, tag.symbol_name)
    if matched_behaviour is None:
      matched_property = name_iteration_helper(sim_vars.integer_properties, tag.symbol_name)
  #Go through all behaviour and properties if no match is found from the above type-specific checks
  if not any((matched_dof, matched_behaviour, matched_property)): 
    matched_behaviour = name_iteration_helper(sim_vars.all_behaviours, tag.symbol_name)
    if matched_behaviour is None:
      matched_property = name_iteration_helper(sim_vars.all_properties, tag.symbol_name)
  #See if any matches are found and save the match to a pair
  if matched_dof is not None:
    pair.pair_type = "dof"
    pair.component_object = matched_dof.Controller.Component
    pair.name = matched_dof.Controller.Component.Name
    pair.behaviour = matched_dof.Controller.Name
    pair.dof = matched_dof.Name
    #Based on example files let's assume that property name = dof name. Could also be vcJoint vs vcDof
    pair.property = matched_dof.Name
  elif matched_behaviour is not None:
    pair.pair_type = "beh"
    pair.component_object = matched_behaviour.Component
    pair.name = matched_behaviour.Component.Name
    pair.behaviour = matched_behaviour.Name
    pair.property = NIL_TEMPLATE #Let's see if this is legit or if we need to change the template
  elif matched_property is not None:
    pair.pair_type = "prop"
    pair.component_object = matched_property.Component
    pair.name = matched_property.Component.Name
    pair.property = matched_property.Name
    pair.behaviour = NIL_TEMPLATE
  return pair


def populate_configuration_info(template_text, server_ip, browsing_tags_text, tags, sim_to_server_text , server_to_sim_text):
  '''Formats template configuration base file with IP address and tag info'''
  template_text = template_text.replace("{IP_ADDRESS_HERE}", server_ip)
  template_text = template_text.replace("{BROWSING_TAGS_HERE}", browsing_tags_text)
  print("-"*50)
  print("Added {} PLC variables to the configuration file.".format(len(tags)))
  template_text = template_text.replace("{SIM_TO_SERVER_HERE}", sim_to_server_text)
  template_text = template_text.replace("{SERVER_TO_SIM_HERE}", server_to_sim_text)
  return template_text


def write_text_to_file(configuration_text):
  '''Writes the populated string to an XML file'''
  uri_xml = os.path.join(base, "ServerConfiguration.xml")
  uri_xml = uri_xml[8:].decode('utf8')

  try:
    with open (uri_xml, "wb") as outfile:
      outfile.write(configuration_text.encode('utf8'))
  except Exception, e:
    print("Failed to save ServerConfiguration.xml. Make sure that the file is not open in another program.")
    return False
  print "-"*50
  print "Configuration file saved to: \n", uri_xml
  return True


def dof_search_helper(sim_vars, symbol_name):
  '''Helper function to attempt to match a real type tag to a simulation DoF'''
  matched_dof = name_iteration_helper(sim_vars.all_dofs, symbol_name)
  if matched_dof is not None:
    return matched_dof
  matched_joint = name_iteration_helper(sim_vars.all_joints, symbol_name)
  if matched_joint is not None:
    if matched_joint.Dof is not None:
      return matched_joint.Dof
  #try to guess the correct dof by splitting the component name
  split_result = symbol_name.split("_M")
  component_name_guess = split_result[0]
  dof_name_guess = "M" + split_result[-1]
  if component_name_guess[-1].isalpha(): #Remove the extra letter (module specifier) from the end
    component_name_guess = component_name_guess[:-1]
  matched_component = name_iteration_helper(sim_vars.all_comps, component_name_guess)
  if matched_component is None:
    component_name_guess = component_name_guess.replace("_", " ")
    matched_component = name_iteration_helper(sim_vars.all_comps, component_name_guess)
  if matched_component is None:
    return None
  #print "Found component that matches a guess:", matched_component.Name
  servos = matched_component.findBehavioursByType(VC_SERVOCONTROLLER)
  for s in servos:
    for joint in s.Joints:
      if joint.Dof is not None:
        #if joint.Dof.Name.endswith(dof_name_guess):
        if dof_name_guess in joint.Dof.Name:
          return joint.Dof
        #if joint.Name.endswith(dof_name_guess):
        if dof_name_guess in joint.Name:
          return joint.Dof
  #print "No match found for", matched_component.Name
  return None


def name_iteration_helper(iterator, symbol_name):
  '''Iterates through a list of behaviours or proeprties and sees if any elements name matches the given symbol name'''
  for x in iterator:
    if x.Name == symbol_name:
      return x
  return None


def initialize_pair_infos(pair_input_data):
  '''Convert the input data list to a list of PairInfos'''
  #pair_input_data = List with row data from Excel
  is_first_row = True
  header_row = []
  pairs = []
  for row in pair_input_data:
    if is_first_row:
      is_first_row = False
      #header_row = row
      continue
    pairs.append(PairInfo(*row))
  return pairs


def attempt_variable_pairing(pair_data):
  '''Read list of rows (ideal pairing formula from Excel) and return list of suitable variable pairs'''
  #process each pair in Pairs
  value_item_pairs = []
  for pair_template in pair_data:
    #Validate if the pair data has proper format
    template_ok = validate_pair_template(pair_template)
    if not template_ok:
      continue
    #Check if the pair has any {index templates}
    component_indexes = extract_text_between_curly_braces(pair_template.ComponentName)
    behaviour_indexes = extract_text_between_curly_braces(pair_template.BehaviourName)
    property_indexes = extract_text_between_curly_braces(pair_template.PropertyName)
    #print pair_template.ComponentName, component_indexes, behaviour_indexes, property_indexes
    #No index template, so this pair template should map 1-to-1
    if not (component_indexes or behaviour_indexes or property_indexes):
      #Check if a matching simulation component name can be found
      sim_components = app.findComponent(pair_template.ComponentName)
      if sim_components is None:
        print("Could not find matching component with name: {}".format(pair_template.ComponentName))
        continue
      value_item_pairs.append(pair_template)
      continue
    #{index template} found. This pair template will map to multiple multiple value item pairs
    #Determine how many pairs should be created. Based on available simulation components
    pairs = pairs_from_template(pair_template, component_indexes, behaviour_indexes, property_indexes)
    if pairs:
      value_item_pairs.extend(pairs)
  return value_item_pairs


def validate_pair_template(pair_template):
  #Return True if the given pair template matches formatting rules (No unnecessary empty fields. Fields have suitable values, etc.)
  VALID_PLUGINS = ["ads"]
  VALID_DIRECTIONS = ["simulation to server", "server to simulation"]
  test_for_empties = [pair_template.Plugin, pair_template.Port, pair_template.Direction, pair_template.ComponentName, pair_template.ServerTagName]
  contains_empty_strings = any(len(text) == 0 for text in test_for_empties)
  if contains_empty_strings:
    return False
  if pair_template.Plugin.lower() not in VALID_PLUGINS:
    return False
  if pair_template.Direction.lower() not in VALID_DIRECTIONS:
    return False
  return True


def pairs_from_template(pair_template, component_indexes, behaviour_indexes, property_indexes):
  '''Read in a PairInfo template and attempt to map it to as many variables as are available in the simulation '''
  pairs_to_return = []
  if not app.Components:
    print("Simulation must have components related to the configuration file. Aborting.")
    return pairs_to_return
  print "Creating pairs from template", pair_template.ComponentName
  #Check if The component name contains #. There might be a related component in simulation without # as the first index
  #Works with integers {i}. Also added support for alphabets {a}
  if "#" in pair_template.ComponentName:
    first_component_name = get_substring_before_hashtag(pair_template.ComponentName)
  else:
    first_component_name = get_substring_before_curlies(pair_template.ComponentName)
  components_with_indexes = find_indexed_components(first_component_name, component_indexes, pair_template.ComponentName)
  if not components_with_indexes:
    print("Could not find components that match the name template: {}".format(pair_template.ComponentName))
    return pairs_to_return
  #3 Options for using index templates
  #Index is used for component name
  if not (behaviour_indexes or property_indexes):
    pairs_to_return = pairs_from_component_indexes(components_with_indexes, pair_template)
    return pairs_to_return
  #Index is used for behaviour name
  if behaviour_indexes:
    pairs = pairs_from_indexed_behaviours(components_with_indexes, pair_template)
    pairs_to_return.extend(pairs)
  #Index is used for property name
  if property_indexes:
    pairs = pairs_from_indexed_properties(components_with_indexes, pair_template)
    pairs_to_return.extend(pairs)
  return pairs_to_return


def find_indexed_components(first_component_name, component_indexes, component_name_template):
  '''Try to look for any components with similar name as the template and extract the indexes'''
  components_with_indexes = []
  one_found = False
  first_component = None
  if component_indexes:
    component_alternatives = [c for c in app.Components if c.Name.startswith(first_component_name)]
    check_integer_index = False
    for ci in component_indexes:
      if "i" in ci:
        check_integer_index = True
        break
    if check_integer_index: #Look for integers i
      for c in component_alternatives:
        index = get_substring_after_hashtag(c.Name)
        if index == "1" and not one_found:
          one_found = True
          first_component = c
        if index != "":
          components_with_indexes.append((c, index))
        elif first_component is None and not one_found:
          first_component = c
    else: #Look for alphabets a
      for c in component_alternatives:
        for letter in ascii_letters:
          name_guess = component_name_template.replace("{a}", letter)
          if c.Name == name_guess:
            components_with_indexes.append((c, letter))
            one_found = True
  if not one_found:
    #one_found refers to the "first" component, as that typically does not have a # in the name
    #The following copied components are typically named ... #2, ... #3, etc.
    first_component = app.findComponent(first_component_name)
    if first_component is not None:
      components_with_indexes.append((first_component, "1"))
  return components_with_indexes


def pairs_from_component_indexes(components_with_indexes, pair_template):
  '''Create pairs based on component name template'''
  pairs = []
  for c, index in components_with_indexes:
    server_tag_name = pair_template.ServerTagName
    # Supported cases: i, i+x, i-x
    if index.isdigit():
      i = int(index)
      server_tag_name = replace_index_template(server_tag_name, i)
    else:
      server_tag_name = server_tag_name.replace("{a}", index)
    pair_info = PairInfo(pair_template.Plugin, pair_template.Address, pair_template.Port, 
                          pair_template.Direction, c.Name, pair_template.BehaviourName, 
                          pair_template.PropertyName, server_tag_name)
    pairs.append(pair_info)
  return pairs


def pairs_from_indexed_behaviours(components_with_indexes, pair_template):
  '''Create pairs based on behaviour name template'''
  pairs = []
  for component, index in components_with_indexes:
    if contains_integer_index(pair_template.BehaviourName): #Check for integer template {i}
      behaviour_prefix = get_substring_before_curlies(pair_template.BehaviourName)
      behaviour_candidates = [x.Name for x in vcHelpers.Selection.getGivenComponentsBehaviours(component) if x.Name.startswith(behaviour_prefix)]
      for i in range(1, len(behaviour_candidates)+1):
        behaviour_name = replace_index_template(pair_template.BehaviourName, i)
        beh = component.findBehaviour(behaviour_name)
        #print prop, property_name
        if beh is None:
          continue
        server_tag_name = replace_index_template(pair_template.ServerTagName, int(index))
        if not index.isdigit():
          server_tag_name = pair_template.ServerTagName.replace("{a}", index)
        pair_info = PairInfo(pair_template.Plugin, pair_template.Address, pair_template.Port, 
                              pair_template.Direction, component.Name, behaviour_name, 
                              pair_template.PropertyName, server_tag_name)
        pairs.append(pair_info)
      else: #Check for alphabets {a}
        for letter in ascii_letters: 
          behaviour_name = pair_template.BehaciourName.replace("{a}", letter)
          if component.findBehaviour(behaviour_name):
            server_tag_name = pair_template.ServerTagName.replace("{a}", letter)
            if index.isdigit():
              server_tag_name = replace_index_template(server_tag_name, int(index))
            pair_info = PairInfo(pair_template.Plugin, pair_template.Address, pair_template.Port, 
                                pair_template.Direction, component.Name, behaviour_name, 
                                pair_template.PropertyName, server_tag_name)
            pairs.append(pair_info)
  return pairs


def pairs_from_indexed_properties(components_with_indexes, pair_template):
  '''Create pairs based on property name template'''
  pairs = []
  for component, index in components_with_indexes:
    if contains_integer_index(pair_template.PropertyName): #Check for integer index {i}
      property_prefix = get_substring_before_curlies(pair_template.PropertyName)
      property_candidates = [x.Name for x in component.Properties if x.Name.startswith(property_prefix)]
      for i in range(1, len(property_candidates)+1):
        property_name = pair_template.PropertyName.replace(".VALUE", "")
        property_name = replace_index_template(property_name, i)
        if not index.isdigit():
          property_name = property_name.replace("{a}", index)
        prop = component.getProperty(property_name)
        #print prop, property_name
        if prop is None:
          continue
        server_tag_name = replace_index_template(pair_template.ServerTagName, i)
        if not index.isdigit():
          server_tag_name = server_tag_name.replace("{a}", index)
        pair_info = PairInfo(pair_template.Plugin, pair_template.Address, pair_template.Port, 
                              pair_template.Direction, component.Name, pair_template.BehaviourName, 
                              property_name, server_tag_name)
        pairs.append(pair_info)
    else: #Check for alphabet {a}
      for letter in ascii_letters: 
        property_name = pair_template.PropertyName.replace("{a}", letter)
        if component.getProperty(property_name):
          server_tag_name = pair_template.ServerTagName.replace("{a}", letter)
          if index.isdigit():
            server_tag_name = replace_index_template(server_tag_name, int(index))
          pair_info = PairInfo(pair_template.Plugin, pair_template.Address, pair_template.Port, 
                              pair_template.Direction, component.Name, pair_template.BehaviourName, 
                              property_name, server_tag_name)
          pairs.append(pair_info)
  return pairs


def contains_integer_index(name):
  for id in INTEGER_INDEX_TEMPLATES:
    if id in name:
      return True
  return False


def replace_index_template(name, i):
  name = name.replace("{i}", str(i))
  name = name.replace("{i+1}", str(i+1))
  name = name.replace("{i-1}", str(i-1))
  return name


def populate_ads_configuration(template_text, ads_pair_info):
  '''Generate pairings text based on available simulation variables'''
  #ads_pair_info = list with PairInfos
  #Extract different servers from the ads_pair_info. Different server = different address/port
  server_infos = set()
  for pair in ads_pair_info:
    server_infos.add((pair.Address, pair.Port))
  #print len(server_infos)
  server_config_text = populate_server_config(server_infos, ads_pair_info)
  template_text = template_text.replace("{SERVERS_HERE}", server_config_text)
  #return the finished xml text that is saved to a file
  return template_text


def populate_server_config(server_infos, ads_pair_info):
  '''Populate ADS address, port, sim to server and server to sim variables'''
  #server_infos = set(Address : string, Port : string)
  server_config_texts = ""
  for i, server_info in enumerate(server_infos, 1): #i starts from 1
    address, port  = server_info
    #print i, address, port
    server_config_text = ADS_SERVER_TEMPLATE
    if address == "":
      address_template = AMS_NET_ID_NILL
    else:
      address_template = AMS_NET_ID_ADDRESS
      address_template = address_template.replace("{ADDRESS_HERE}", address)
    server_config_text = server_config_text.replace("{ADDRESS_TEMPLATE_HERE}", address_template)
    server_config_text = server_config_text.replace("{PORT_HERE}", port)
    if i == 1:
      server_name = "ADS Server"
    else:
      server_name = "ADS Server {}".format(i)
    server_config_text = server_config_text.replace("{SERVER_NAME_HERE}", server_name)
    sim_to_ser, ser_to_sim_c, ser_to_sim_e = populate_variable_groups(ads_pair_info, address, port)
    variable_group_text = populate_variable_group_text(sim_to_ser, ser_to_sim_c, ser_to_sim_e)
    server_config_text = server_config_text.replace("{VARIABLE_GROUPS_HERE}", variable_group_text)
    server_config_texts = server_config_texts + server_config_text
  return server_config_texts


def populate_variable_groups(ads_pair_info, address, port):
  '''Populate the value items for each variable group'''
  sim_to_server_variable_group_text = ""
  server_to_sim_cyclic_variable_group_text = ""
  server_to_sim_event_variable_group_text = ""
  for pair_info in ads_pair_info:
    if pair_info.Address != address or pair_info.Port != port:
      continue
    variable_item_text = populate_variable_item(pair_info)
    #Sim -> Server group should always be cyclic to ensure good performance
    if pair_info.Direction.lower() == "server to simulation":
      if sim_to_server_variable_group_text != "":
        variable_item_text = "\n\t\t\t\t\t\t\t" + variable_item_text
      sim_to_server_variable_group_text = sim_to_server_variable_group_text + variable_item_text
    elif pair_info.Direction.lower() == "simulation to server":
      #Improvement idea is to automatically add Server -> Sim booleans to an event-based group and other variables to an cyclic group
      if True:
        if server_to_sim_cyclic_variable_group_text != "":
          variable_item_text = "\n\t\t\t\t\t\t\t" + variable_item_text
        server_to_sim_cyclic_variable_group_text = server_to_sim_cyclic_variable_group_text + variable_item_text
      else:
        if server_to_sim_event_variable_group_text != "":
          variable_item_text = "\n\t\t\t\t\t\t\t" + variable_item_text
        server_to_sim_event_variable_group_text = server_to_sim_event_variable_group_text + variable_item_text
    else:
      #Invalid. Should never happen
      continue
  return sim_to_server_variable_group_text, server_to_sim_cyclic_variable_group_text, server_to_sim_event_variable_group_text


def populate_variable_group_text(sim_to_server_variable_group_text, server_to_sim_cyclic_variable_group_text, server_to_sim_event_variable_group_text):
  '''Add all the variable groups texts to one string that can be added to the server config'''
  variable_group_text = ""
  sim_to_server_text = ADS_VARIABLE_GROUP_TEMPLATE
  server_to_sim_cyclic_text = ADS_VARIABLE_GROUP_TEMPLATE
  server_to_sim_event_text = ADS_VARIABLE_GROUP_TEMPLATE
  if sim_to_server_variable_group_text != "":
    sim_to_server_text = sim_to_server_text.replace("{DIRECTION_HERE}", "SimulationToServer")
    sim_to_server_text = sim_to_server_text.replace("{GROUP_NAME_HERE}", "Simulation to server")
    sim_to_server_text = sim_to_server_text.replace("{UPDATE_METHOD_HERE}", "Cyclic")
    sim_to_server_text = sim_to_server_text.replace("{VARIABLE_ITEMS_HERE}", sim_to_server_variable_group_text)
    variable_group_text = variable_group_text + sim_to_server_text
  if server_to_sim_cyclic_variable_group_text != "":
    if variable_group_text != "":
      variable_group_text =  variable_group_text + "\n\t\t\t\t\t\t"
    server_to_sim_cyclic_text = server_to_sim_cyclic_text.replace("{DIRECTION_HERE}", "ServerToSimulation")
    server_to_sim_cyclic_text = server_to_sim_cyclic_text.replace("{GROUP_NAME_HERE}", "Server to simulation")
    server_to_sim_cyclic_text = server_to_sim_cyclic_text.replace("{UPDATE_METHOD_HERE}", "Cyclic")
    server_to_sim_cyclic_text = server_to_sim_cyclic_text.replace("{VARIABLE_ITEMS_HERE}", server_to_sim_cyclic_variable_group_text)
    variable_group_text = variable_group_text + server_to_sim_cyclic_text
  if server_to_sim_event_variable_group_text != "":
    if variable_group_text != "":
      variable_group_text = variable_group_text + "\n\t\t\t\t\t\t"
    server_to_sim_event_text = server_to_sim_event_text.replace("{DIRECTION_HERE}", "ServerToSimulation")
    server_to_sim_event_text = server_to_sim_event_text.replace("{GROUP_NAME_HERE}", "Server to simulation event-based")
    server_to_sim_event_text = server_to_sim_event_text.replace("{UPDATE_METHOD_HERE}", "ImmediateEventBased")
    server_to_sim_event_text = server_to_sim_event_text.replace("{VARIABLE_ITEMS_HERE}", server_to_sim_event_variable_group_text)
    variable_group_text = variable_group_text + server_to_sim_event_text
  return variable_group_text


def populate_variable_item(pair_info):
  '''Fill the variable item template with the pair_info'''
  #print pair_info.ComponentName, pair_info.BehaviourName, pair_info.PropertyName
  is_dof_variable = is_dof(pair_info)
  is_signal_map_variable = is_signal_map(pair_info)
  if is_dof_variable:
    variable_item_text = DOF_VARIABLE_TEMPLATE
  elif is_signal_map_variable:
    variable_item_text = SIGNAL_MAP_VARIABLE_TEMPLATE
  else:
    variable_item_text = STANDARD_VARIABLE_TEMPLATE
  variable_item_text = variable_item_text.replace("{ComponentName}", pair_info.ComponentName)
  #Behaviour
  if pair_info.BehaviourName:
    variable_item_text = variable_item_text.replace("{BehaviourTemplate}", BEHAVIOUR_NAME_TEMPLATE)
    variable_item_text = variable_item_text.replace("{BehaviourName}", pair_info.BehaviourName)
  else:
    variable_item_text = variable_item_text.replace("{BehaviourTemplate}", EMPTY_BEHAVIOUR_TEMPLATE)
  #Property
  if pair_info.PropertyName:
    if is_signal_map_variable:
      variable_item_text = variable_item_text.replace("{PropertyTemplate}", EMPTY_PROPERTY_TEMPLATE)
      variable_item_text = variable_item_text.replace("{PortNumber}", pair_info.PropertyName)
    else: #Not a signal map
      variable_item_text = variable_item_text.replace("{PropertyTemplate}", PROPERTY_NAME_TEMPLATE)
      variable_item_text = variable_item_text.replace("{PropertyName}", pair_info.PropertyName)
  else: #Signal behaviours do not have a property
    variable_item_text = variable_item_text.replace("{PropertyTemplate}", EMPTY_PROPERTY_TEMPLATE)
  variable_item_text = variable_item_text.replace("{ServerVariableName}", pair_info.ServerTagName)
  #variable_item_text = variable_item_text + "\n"
  return variable_item_text


def extract_text_between_curly_braces(text):
  # Define the regex pattern to match text between curly braces
  pattern = r"\{(.*?)\}"
  # Find all matches in the input text
  matches = re.findall(pattern, text)
  # Return the list of matched strings
  return matches


def get_substring_before_curlies(text):
  # Find the index of the first curly brace symbol
  index = text.find("{")
  # If first curly brace symbol is not found, return the entire string
  if index == -1:
      return text.strip()
  else:  # Return the substring before the first curly brace symbol
      return text[:index].strip()


def get_substring_before_hashtag(text):
  # Find the index of the first hashtag symbol
  index = text.find("#")
  # If hashtag symbol is not found, return the entire string
  if index == -1:
      return text.strip()
  else: # Return the substring before the hashtag symbol
      return text[:index].strip()


def get_substring_after_hashtag(text):
  # Find the index of the first hashtag symbol
  index = text.find("#")
  # If hashtag symbol is not found, return the entire string
  if index == -1:
      return ""
  else: # Return the substring after the hashtag symbol
      return text[index+1:].strip()


def is_dof(pair_info):
  '''Return True if the given pair contains a moving joint / DoF'''
  if "VALUE" in pair_info.PropertyName:
    return True
  comp = app.findComponent(pair_info.ComponentName)
  if comp is None:
    return False
  beh = comp.findBehaviour(pair_info.BehaviourName)
  if beh is None:
    return False
  if beh.Type in [VC_SERVOCONTROLLER, VC_ROBOTCONTROLLER]:
    return True
  return False


def is_signal_map(pair_info):
  '''Return True if the given pair contains a signal map behaviour'''
  if not pair_info.PropertyName.isdigit(): #Return False if the property is column does not have an integer digit (port index)
    return False
  comp = app.findComponent(pair_info.ComponentName)
  if comp is None:
    return False
  beh = comp.findBehaviour(pair_info.BehaviourName)
  if beh is None:
    return False
  if beh.Type == VC_BOOLEANSIGNALMAP:
    return True
  return False


def load_config():
  '''Launch the Import Configuration window'''
  uri_xml = os.path.join(base, "ServerConfiguration.xml")
  uri_xml = uri_xml[8:].decode('utf8')
  cmd = app.findCommand("netCommand")
  cmd.execute("ConConfigurationImport")

#endregion

#region Class definitions
#Class for organizing tag information
class Tag:
  def __init__(self):
    self.memory_location = None
    self.bit_index = None
    self.byte_index = None
    self.data_area = None
    self.data_type = None
    self.symbol_name = None
    self.word_length = None
    self.server_data_type = None

#Class for organizing component information for creating varible pairs
class PairedComponent:
  def __init__(self):
    self.pair_type = None
    self.component_object = None
    self.name = None
    self.behaviour = None
    self.property = None
    self.dof = None

#Class for organizing all simulation variables by type
class SimulationVariables:
  def __init__(self):
    self.all_properties = []
    self.all_comps = vcHelpers.Selection.getAllComponents()
    for c in self.all_comps:
      self.all_properties.extend(c.Properties)
    self.all_behaviours = vcHelpers.Selection.getAllComponentsBehaviours()
    #
    self.all_servos = vcHelpers.Selection.filterBehaviours(self.all_behaviours, VC_SERVOCONTROLLER)
    self.all_joints = []
    for s in self.all_servos:
      self.all_joints.extend(s.Joints)
    self.all_dofs = [x.Dof for x in self.all_joints]
    #
    self.boolean_signals = vcHelpers.Selection.filterBehaviours(self.all_behaviours, VC_BOOLEANSIGNAL)
    self.boolean_properties = [p for p in self.all_properties if p.Type == VC_BOOLEAN]
    #Boolean maps not included. Any boolean map signals that are already connected already have a hidden boolean signal behind an interface
    self.integer_signals = vcHelpers.Selection.filterBehaviours(self.all_behaviours, VC_INTEGERSIGNAL)
    self.integer_properties = [p for p in self.all_properties if p.Type == VC_INTEGER]
    self.real_signals = vcHelpers.Selection.filterBehaviours(self.all_behaviours, VC_REALSIGNAL)
    self.real_properties = [p for p in self.all_properties if p.Type == VC_REAL]
    self.string_signals = vcHelpers.Selection.filterBehaviours(self.all_behaviours, VC_STRINGSIGNAL)
    self.string_properties = [p for p in self.all_properties if p.Type == VC_STRING]
    self.matrix_signals = vcHelpers.Selection.filterBehaviours(self.all_behaviours, VC_MATRIXSIGNAL)
    self.matrix_properties = [p for p in self.all_properties if p.Type == VC_MATRIX]
    self.vector_properties = [p for p in self.all_properties if p.Type == VC_VECTOR]

#for ADS pairing 
class PairInfo():
  def __init__(self, plugin, address, port, direction, component, beh, prop, tag):
    self.Plugin = plugin
    self.Address = address
    self.Port = port
    self.Direction = direction
    self.ComponentName = component
    self.BehaviourName = beh
    self.PropertyName = prop
    self.ServerTagName = tag


#endregion

app = getApplication()
addState(first_state)
